/* Level 6 theremin
    This program plays songs presented by:
        https://dragaosemchama.com/en/2019/02/songs-for-arduino/
   It uses DAC circuit described by Amanda Ghassaei
        https://www.instructables.com/id/Arduino-Audio-Output/
    Features
      Music: yes
      6 bit DAC on PORTB (pins 8-13): yes
      8 bit DAC on PORTB (pins 0-7): not supported
      tremolo: yes
      Note bending: yes
      Fade-in and out at beginning of notes to reduce pops: yes
  Kubichek. Sept 2012
*/
//#define DEBUG       //comment this line for actual use
#include "pitches.h"  //copied from Example program "toneMelody" in Digital group.
#define NumTableEntries 80  //Make smaller if long song runs you short of memory
//                          output gets noisier for values less than 70-80
#define SampleRate 16130  //sample rate. see set_interrupts()
#define alpha .993       //closer to 1 bends the note longer

//User can turn on/off tremolo and bend as desired:
volatile boolean tremolo = false;   //Set this to false to turn off tremolo effect
boolean bend = false;                //Set to true to bend the notes
#define LDRpin A3                   //photoresistor pin
#define PUSHpin A4                  //pushbutton switch
#define LEDpin A5                   //LED
int lastSwitch = HIGH;              //last pushbutton state
//User can experiment with different values of tremSweep and tremFreq
int tremSweep = 20; //max freq deviation from desiredFreq. Hz (suggest 20)
int tremFreq = 7;   //controls how fast the waa-waa changes. Hz (suggest 7)

//User should not fool with following.
volatile unsigned int desiredFreq = 1000;
byte SineTable[NumTableEntries];    //sine table scaled to 8 bits
int SineTabTrem[NumTableEntries];  //scaled for tremolo shift. can be negative
float ramp[] = {0, .1, .2, .3, .4, .5, .6, .7, .8, .9, 1.0};
int Nramp = 11;
volatile boolean soundOn = false;  //use to turn on/off output to DAC
float deltaTab;
float avgFrq;                     //used for bending notes
#define REST      0               //apparently used in the song .h files

//                    ---CHOOSE A SONG HERE---

// notes of the moledy followed by the duration.
//    a 4 means a quarter note, 8 an eighteenth , 16 sixteenth, so on
//    !!negative numbers are used to represent dotted notes,
//    so -4 means a dotted quarter note, that is, a quarter plus an eighteenth!!

//Following melody arrays copied from: https://dragaosemchama.com/en/2019/02/songs-for-arduino/
//    Uncomment both a tempo and an include.  Set tempo as desired
//int tempo = 100;
//#include "lullaby.h"          //suggest tempo 80 to 100
//int tempo = 100;
//#include "starwars.h"         //suggest at least tempo = 100
//int tempo = 150;
//#include "merryChristmas.h"   //suggest tempo = 150
//int tempo = 100;
//#include "silentNight.h"        //suggest tempo = 100
//int tempo = 150;
//#include "lionSleeps.h"       //suggest tempo = 150
int tempo = 80;
#include "greensleeves.h"     //suggest tempo=80
//int tempo = 100;
//#include "darth.h"            //tempo = 100 is ok

int notes = sizeof(melody) / sizeof(melody[0]) / 2;
int wholenote = (60000 * 4) / tempo; //duration of whole note in ms

void set_interrupts() {
  //This routine sets DAC up using timer2 interrupt at ~16 kHz.
  //We don't use timer 0 since it is used by millis()and delay. Also, timer 1 is used for pwm
  //Based on amandaghassaei: https://www.instructables.com/Arduino-Timer-Interrupts/
  cli();//disable interrupts
  TCCR2A = 0;// set entire TCCR2A register to 0
  TCCR2B = 0;// same for TCCR2B
  TCNT2  = 0;//initialize counter value to 0
  //          set compare match register for 16khz increments
  OCR2A = 124;// = (16*10^6) / (16000*8) - 1 (must be <256)
  //          actual freq = (16*10^6)/(prescalar*cmr)+1=16*10^6/(8*124)+1 = 16130
  TCCR2A |= (1 << WGM21);   // turn on CTC mode
  TCCR2B |= (1 << CS21);    // Set CS21 bit for 8 prescaler
  TIMSK2 |= (1 << OCIE2A);  // enable timer compare interrupt
  sei();  //enable interrupts
}
void FillSineTable() { //precompute sine tables to speed things up
  for (int i = 0; i < NumTableEntries; i++) { // fill table values
    float S = sin(i * (float)(2 * PI / NumTableEntries));
    SineTable[i] = 127 * S + 127; //scale to 8 bits
    SineTabTrem[i] = tremSweep * S; //scale to tremolo freq shift
  }
}
ISR(TIMER2_COMPA_vect) { //timer 2 interrupt
  /*
     tremolo: optionally, a small sinusoidally changing increment is added to desiredFreq.
        It causes a "waa-waa" sound. The rate of waa-waa is given by tremFreq.
        The max tremolo freq deviation is given by tremSweep. See FillSineTable
     When sound is turned on or off, there is a small pop. To reduce this,
        apply a ramp function: When sound switches off, decay amplitude to 127 (center value)
        When sound switches on, ramp amplitudes from 127 to the next sound amplitude.
  */
  static unsigned int tremindx = 0;
  static unsigned int index = 0;
  static boolean lastSound = false;
  static boolean rampOn = false;
  static unsigned int nr = 0;
  int FreqAdj = 0;
  byte out;

  //tremolo effect
  if (tremolo) {
    tremindx += tremFreq;    // calculate the next phase
    if (tremindx >= SampleRate)   // keep phase between 0-2*pi
      tremindx -= SampleRate;
    unsigned int indxFrq = (tremindx * deltaTab);  //scale pointer to table size
    FreqAdj = SineTabTrem[indxFrq];  //adjust dsrd freq up or down
    if (FreqAdj + desiredFreq < 0) FreqAdj = 0; //in case desiredFreq = 0
  }
  //main freq pointer
  index += desiredFreq + FreqAdj;  // calculate the next phase
  if (index >= SampleRate)  // keep phase between 0-2*pi
    index -= SampleRate;
  int findx = (index * deltaTab);    //scale pointer to table size
  out = SineTable[findx];

  //Now apply the ramp function at transitions
  if (soundOn != lastSound) { //detect on-off or off-on transition
    rampOn = true;
    nr = 0;
  }
  lastSound = soundOn;
  if (rampOn) { //apply ramp to reduce pops
    if (soundOn) out = (out - 127) * ramp[nr++] + 127; //taper on ramp
    else out = (out - 127) * (1.0 - ramp[nr++]) + 127; //taper off ramp
    if (nr >= Nramp) rampOn = false;
  }

#ifdef DEBUG
  //Serial.println(out, BIN); //In interrupt, serial print only works for BIN OCT or HEX
#else
  //output if soundOn or if ramp still active
  if (soundOn || rampOn) {
    //if (soundOn){
    PORTB = out >> 2;
  }
  else PORTB = 31;
#endif
}
void checktremolo() {
  //turn tremolo on or off
  int nowSwitch = digitalRead(PUSHpin); //tremolo on or off?
  if (lastSwitch == HIGH && nowSwitch == LOW) {
    tremolo = !tremolo;
    digitalWrite(LEDpin, tremolo);
  }
  lastSwitch = nowSwitch;
}
void setup() {
#ifdef DEBUG
  Serial.begin(9600);
#endif
  for (int i = 8; i < 14; i++) pinMode(i, OUTPUT);  //PORTB pins: 8-13
  set_interrupts();
  FillSineTable();
  deltaTab = (float) NumTableEntries / (float) SampleRate; //used for scaling
  pinMode(PUSHpin, INPUT_PULLUP); //for changing options
  pinMode(LEDpin, OUTPUT);
  digitalWrite(LEDpin, tremolo);
}
int divider, noteDuration;
void loop() {
  // iterate over the notes of the melody.
  // Remember, the array is twice the number of notes (notes + durations)
  for (int thisNote = 0; thisNote < notes * 2; thisNote = thisNote + 2) {
    // calculate the duration of each note
    divider = melody[thisNote + 1];
    if (divider > 0) { // regular note, just proceed
      noteDuration = (wholenote) / divider;
    } else if (divider < 0) { // dotted notes represented by negative durations!!
      noteDuration = (wholenote) / abs(divider);
      noteDuration *= 1.5; // increases the duration in half for dotted notes
    }
    soundOn = true;  //turns on output in the ISR
    unsigned long T0 = millis();
    while (millis() - T0 < noteDuration) { //hold note on for desired delay.
      checktremolo(); //see if tremolo on or off
      if (bend) {  //bend note by running average of frequencies
        delay(1); //makes it easier to bend the note
        if (thisNote == 0)avgFrq = melody[thisNote];
        else avgFrq = alpha * avgFrq + (1 - alpha) * melody[thisNote];
        desiredFreq = avgFrq;
      }
      else desiredFreq = melody[thisNote];  //no note bending
    }
    // short pause between notes.
    soundOn = false;  //turn off sound output in ISR
    delay(noteDuration * .2);
  }
  delay(1000); //wait a sec before repeating the song
}
